/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>

#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>

#include "common.h"
#include "log.h"

char *hex(void *s,size_t len) {
	char *str;
	size_t i;

	if ((str = malloc((len*2)+1)) == NULL)
		return NULL;

	bzero(str,(len*2)+1);

	for (i = 0; i < len; i++) {
		sprintf(&str[i*2],"%.2x",(unsigned char)((char *)s)[i]);
	}

	return str;
}

#ifndef HAVE_INET_PTON
int inet_pton(int af,const char *src, void *dst) {
	struct in_addr addr;

	/* if (af != AF_INET) abort(); */

	if (inet_aton(src,&addr)) {
		memcpy(dst,&addr,sizeof(struct in_addr));
		return 1;
	}
	return 0;
}
#endif


int sopen(const char* path,...) {
	va_list args;
	struct stat st;
	int sd;
	struct sockaddr_un addr;
	
	if (stat(path,&st) == -1)
		return -1;

	if (!S_ISSOCK(st.st_mode)) {
		errno = EINVAL;
		return -1;
	}
	
	if ((sd = socket(PF_LOCAL,SOCK_STREAM,0)) == -1)
		return -1;

	bzero(&addr,sizeof(addr));

	addr.sun_family = AF_LOCAL;
	strncpy(addr.sun_path,path,sizeof(addr.sun_path)); /* path has nul */

	if (connect(sd,(struct sockaddr *)&addr,sizeof(addr)) == -1)
		return -1;

	return sd;
}


void *_malloc(size_t s) {
	void *p;
	if ((p = malloc(s)) != NULL)
		DUMP("allocated %i bytes at pointer %x",s,(intptr_t)p);
	return p;
}

void _free(void *p) {
	DUMP("freed pointer %x",(intptr_t)p);
	free(p);
}
