/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>

#include <arpa/inet.h>

#include <gmp.h>

#include "common.h"
#include "object.h"
#include "thread.h"
#include "log.h"
#include "packet.h"
#include "queue.h"
#include "link.h"
#include "crypt.h"
#include "bignum.h"


void link_mixmaster() {
	struct timespec sleep;	/* nanosleep	*/
	struct timeval start, finish;	/* gettimeofday */
	link_ctx *l;
	prand_ctx r;
	unsigned char entropy[MAX(255,sizeof(link_list.list))]	= {0};
	unsigned char arc4_stream[PACKET_SIZE]	= {0};
	unsigned char *null_pk[PACKET_SIZE]	= {0};
	SLIST_HEAD(,link_ctx) closeq;

	object_init(&link_list,(void(*)(void *))object_destroy,&link_list);
	object_retain(&link_list);

	rand_get(entropy,MIN(255,MAX(255,sizeof(entropy))));
	prand_init(&r,ARCFOUR_STREAM,entropy,MIN(255,MAX(255,sizeof(entropy))));

	sleep.tv_sec	= 0;
	sleep.tv_nsec	= 0;
	start.tv_sec	= 0;
	start.tv_usec	= 0;
	finish.tv_sec	= 0;
	finish.tv_usec	= 0;

	for (;;) {
		int i;
		
		gettimeofday(&start,NULL);

		object_lock(&link_list);

		/*
		 * write out buffer
		 */
		SHUFFLE_LINK_LIST(&r);

		for (i = 0; i < link_list.length; i++) {
			l = link_list.list[i];
			link_try_write(l);
		}

		/*
		 * read in buffer
		 */
		SHUFFLE_LINK_LIST(&r);

		for (i = 0; i < link_list.length; i++) {
			l = link_list.list[i];
			link_try_read(l);
		}


		/*
		 * close failed sockets
		 *
		 * !!! just temporary... this is *not* how
		 * !!! it should be done
		 */
		SHUFFLE_LINK_LIST(&r);
		SLIST_INIT(&closeq);
		for (i = 0; i < link_list.length; i++) {
			l = link_list.list[i];
			object_lock(l);
				if (l->failed) {
					SLIST_INSERT_HEAD(&closeq,l,sle);
				}
			object_unlock(l);
		}
		while ((l = SLIST_FIRST(&closeq)) != NULL) {
			SLIST_REMOVE_HEAD(&closeq,sle);
			close(l->sd);
			link_unregister_nm(l);
		}


		object_unlock(&link_list);

		gettimeofday(&finish,NULL);
		
		if ((sleep.tv_nsec = LINK_NANO_DELAY - ((finish.tv_usec - start.tv_usec) * 1000)) < LINK_NANOSLEEP_COST)
			continue;

		nanosleep(&sleep,NULL);

	}

	object_release(&link_list);
	prand_destroy(&r);
} /* link_mixmaster */


void link_try_read(link_ctx *l) {
	ssize_t in, wrote;
	packet_t *packet;

	object_lock(l);

	if ((packet = SLIST_FIRST(&l->cache)) != NULL) {
		if ((in = recv(l->sd,packet->data,PACKET_SIZE,MSG_PEEK)) >= PACKET_SIZE) {
			if ((in = read(l->sd,packet->data,PACKET_SIZE)) == PACKET_SIZE) {
				prand_xor(&l->arc4_in,packet->data,PACKET_SIZE);
				if (FLAG_ISSET(packet->data[0],PACKET_LINK_NULL)) {
					PACKET_CLEAN(packet);
					if (!l->failed) l->in_null++;
				} else {
					l->in_null = 0;
					SLIST_REMOVE_HEAD(&l->cache,sle);
					l->cache_len--;
					SIMPLEQ_INSERT_TAIL(&l->in,packet,sqe);
					l->in_len++;
				}
			} else {
				/*** socket closed on us ***/
				l->failed = true;
			}
		} else if (in == -1) {
			switch (errno) {
				#if defined EAGAIN
				case EAGAIN:
				#elif defined EWOULDBLOCK
				case EWOULDBLOCK:
				#else
				#error No non-blocking status available
				#endif
				/*** empty ***/
				break;
				default:
					/*** socket closed on us ***/
					l->failed = true;
			}
		} else {
			/*** not enough data to read ***/
		}
	} else { /*** cache empty!? ***/
		ERROR("no packets available in cache!");
		l->failed = true;
	}

	if (l->in_null > LINK_MAX_NULL)
		l->failed = true;

	if ((l->in_len > 0) || l->failed)
		object_broadcast(l);

	object_unlock(l);


} /* link_try_read */


void link_try_write(link_ctx *l) {
	ssize_t out, wrote;
	packet_t *packet;

	object_lock(l);

	try_write: if ((packet = SIMPLEQ_FIRST(&l->out)) != NULL) {
		if (!packet->encrypted) {
			prand_xor(&l->arc4_out,packet->data,PACKET_SIZE);
			packet->encrypted = true;
		}

		if ((out = write(l->sd,packet->pos,(PACKET_SIZE - (packet->pos - packet->data)))) > 0) {
			if (out == (PACKET_SIZE - (packet->pos - packet->data))) {
				if (packet->isnull) {
					if (!l->failed) l->out_null++;
				} else {
					l->out_null = 0;
				}
				SIMPLEQ_REMOVE_HEAD(&l->out,packet,sqe);
				l->out_len--;
				PACKET_CLEAN(packet);
				SLIST_INSERT_HEAD(&l->cache,packet,sle);
				l->cache_len++;
			} else {
				packet->pos += out;
			}
		} else if (out == -1) {
			switch (errno) {
				#if defined EAGAIN
				case EAGAIN:
				#elif defined EWOULDBLOCK
				case EWOULDBLOCK:
				#else
				#error No non-blocking status available
				#endif
				/*** buffer full ***/
				break;
				default:
					/*** socket closed on us ***/
					l->failed = true;
			}
		} else { /* socket closed */
			l->failed = true;
			l->closed = true;
		}
	} else { /*** nothing to send, construct null packet ***/
		if ((packet = SLIST_FIRST(&l->cache)) != NULL) {
			SLIST_REMOVE_HEAD(&l->cache,sle);
			l->cache_len--;
			FLAG_XSET(packet->data[0],PACKET_LINK_NULL);
			packet->isnull = true;
			SIMPLEQ_INSERT_TAIL(&l->out,packet,sqe);
			l->out_len++;
			goto try_write;
		} else {
			ERROR("no packets available in cache!");
			l->failed = true;
		}
	}

	if (l->out_null > LINK_MAX_NULL)
		l->failed = true;

	if (l->failed)
		object_broadcast(l);

	object_unlock(l);

} /* link_try_write */


void link_register(link_ctx *l) {
	object_lock(&link_list);
		if (link_list.length < (int)sizeof(link_list.list)) {
			object_retain(l);
			link_list.list[link_list.length] = l;
			link_list.length++;
		} else
			ERROR("cannot register link %x: queue overflow",l->link_id);
	object_unlock(&link_list);
} /* link_register */


void link_unregister(link_ctx *l) {
	
	object_lock(&link_list);

	link_unregister_nm(l);

	object_unlock(&link_list);

} /* link_unregister */

void link_unregister_nm(link_ctx *l) {
	int i;
	bool found = false;
	
	for (i = 0;i < link_list.length;i++) {
		if (link_list.list[i] == l) {
			SWAP(link_ctx *,link_list.list[i],link_list.list[(link_list.length-1)]);
			link_list.length--;
			found = true;
			object_release(l);
			break;	
		}
	}

	if (!found)
		ERROR("could not unregister link %x: not in queue",l->link_id);
} /* link_unregister */


void link_setsockopts(int s) {
	int opt;

	/*
	 * SO_RCVLOWAT & SO_SNDLOWAT not portable :(
	 * 
	 * opt = PACKET_SIZE;
	 * if (setsockopt(s,SOL_SOCKET,SO_RCVLOWAT,&opt,sizeof(opt)) != 0)
	 * 	WARN("could not set SO_RCVLOWAT on link socket: %s",strerror(errno));
	 * 
	 * opt = PACKET_SIZE * 20;
	 * if (setsockopt(s,SOL_SOCKET,SO_SNDLOWAT,&opt,sizeof(opt)) != 0)
	 * 	WARN("could not set SO_SNDLOWAT on link socket: %s",strerror(errno));
	 */ 

	opt = IPTOS_LOWDELAY;
	if (setsockopt(s,IPPROTO_IP,IP_TOS,&opt,sizeof(opt)) != 0)
		WARN("could not set IP_TOS on link socket: %s",strerror(errno));

} /* link_setsockopts */


link_ctx *link_create(int sd) {
	link_ctx *l;
	int i, opts;
	
	if ((l = _malloc(sizeof(link_ctx))) == NULL) {
		ERROR("could not malloc space for a new link");
		return NULL;
	}
	bzero(l,sizeof(link_ctx));

	object_init(l,link_free,l);

	l->link_id = (long int)l;
	l->sd = sd;

	/* set non-blocking io */
	opts = fcntl(sd,F_GETFL,0);
	fcntl(sd,F_SETFL,opts | O_NONBLOCK);

	/* initialize packet lists/queues */
	SLIST_INIT(&l->cache);
	SIMPLEQ_INIT(&l->in);
	SIMPLEQ_INIT(&l->out);

	/* add the packets from buffer into cache queue */
	for (i = 0; i < PACKET_CACHE_MAX; i++) {
		l->packet_alloc[i].link_id	= l->link_id;
		l->packet_alloc[i].pos		= l->packet_alloc[i].data;
		SLIST_INSERT_HEAD(&l->cache,&l->packet_alloc[i],sle);
	}

	l->cache_len	= PACKET_CACHE_MAX;
	l->in_len	= 0;
	l->out_len	= 0;
	l->in_null	= 0;

	l->iscaller	= false;
	l->initiated	= false;
	l->failed	= false;
	l->closed	= false;

	__objects_created++;

	return l;
} /* link_create */


void link_free(void *l) {

	object_destroy(l);

	if (((link_ctx *)l)->initiated) {
		prand_destroy(&((link_ctx *)l)->kdf);
		prand_destroy(&((link_ctx *)l)->arc4_in);
		prand_destroy(&((link_ctx *)l)->arc4_out);
	}


	_free(l);

	__objects_freed++;
	
	return; /* void */
} /* link_free */


bool link_do_read(link_ctx *l,void *buf,size_t len,bool cipher) {
	unsigned char pkt[PACKET_SIZE]	= {0};
	unsigned char *b		= (unsigned char *)buf;
	int n				= (int)MIN(len,INT_MAX);
	unsigned char f			= (l->initiated)?
						PACKET_LINK_NBND :
						PACKET_LINK_INIT;
	/*
	 * read buffer
	 */
	while (n > 0) {
		unsigned char *p;
		int in, c;
		fd_set fdset;
		struct timeval tv;

		/*
		 * read packet
		 */
		p = pkt;
		c = PACKET_SIZE;
		while (c > 0) {
			try_read: if ((in = read(l->sd,p,c)) > 0) {
				p += in;
				c -= in;
				continue;
			} else if (in == -1) {switch (errno) {
				#if defined EAGAIN
				case EAGAIN:
				#elif defined EWOULDBLOCK
				case EWOULDBLOCK:
				#else
				#error No non-blocking status available
				#endif
					FD_ZERO(&fdset);
					FD_SET(l->sd,&fdset);
					tv.tv_sec	= 1;
					tv.tv_usec	= 0;
					switch(select(l->sd+1,&fdset,NULL,&fdset,&tv)) {
						case 0:
							INFO("timeout on socket");
							return false;
						case -1:
							INFO("error waiting on socket");
							return false;
						default:
							goto try_read;
					}
				default:
					INFO("error reading socket");
					return false;
			}} else {
				INFO("socket closed unexpectedly");
				return false;
			}
		}

		if (cipher)
			prand_xor(&l->arc4_in,pkt,PACKET_SIZE);

		if (!FLAG_ISXSET(*pkt,f)) {
			WARN("packet header not set correctly");
			return false;
		}

		memcpy(b,pkt+1,MIN(n,(PACKET_SIZE-1)));
		b	+= (PACKET_SIZE-1);
		n	-= (PACKET_SIZE-1);
	}

	return true;
} /* link_do_read */


bool link_do_write(link_ctx *l,void *buf,size_t len,bool cipher) {
	unsigned char pkt[PACKET_SIZE]	= {0};
	unsigned char *b		= (unsigned char *)buf;
	int n				= (int)MIN(len,INT_MAX);
	unsigned char f			= (l->initiated)?
						PACKET_LINK_NBND :
						PACKET_LINK_INIT;

	while (n > 0) {
		unsigned char *p;
		int out, c;
		fd_set fdset;
		struct timeval tv;

		FLAG_XSET(*pkt,f);
		memcpy(pkt+1,b,MIN(n,(PACKET_SIZE-1)));
		
		if (cipher)
			prand_xor(&l->arc4_out,pkt,PACKET_SIZE);

		/*
		 * write packet
		 */
		p = pkt;
		c = PACKET_SIZE;
		while (c > 0) {
			try_read: if ((out = write(l->sd,p,c)) > 0) {
				p += out;
				c -= out;
				continue;
			} else if (out == -1) {switch (errno) {
				#if defined EAGAIN
				case EAGAIN:
				#elif defined EWOULDBLOCK
				case EWOULDBLOCK:
				#else
				#error No non-blocking status available
				#endif
					FD_ZERO(&fdset);
					FD_SET(l->sd,&fdset);
					tv.tv_sec	= 1;
					tv.tv_usec	= 0;
					switch(select(l->sd+1,NULL,&fdset,&fdset,&tv)) {
						case 0:
							INFO("timeout on socket");
							return false;
						case -1:
							INFO("error waiting on socket");
							return false;
						default:
							goto try_read;
					}
				default:
					INFO("error writing socket");
					return false;
			}} else {
				INFO("socket closed unexpectedly");
				return false;
			}
		}

		b	+= (PACKET_SIZE-1);
		n	-= (PACKET_SIZE-1);
	}

	return true;
} /* link_do_write */


packet_t *link_cache_get(link_ctx *l) {
	packet_t *p;
	
	object_lock(l);
		if ((p = SLIST_FIRST(&l->cache)) != NULL) {
			SLIST_REMOVE_HEAD(&l->cache,sle);
			l->cache_len--;
		}
	object_unlock(l);

	return p;
} /* link_cache_get */


void link_cache_put(link_ctx *l,packet_t *p) {
	ASSERT(p != NULL,"null packet passed to %s()",__func__);

	object_lock(l);
		PACKET_CLEAN(p);
		SLIST_INSERT_HEAD(&l->cache,p,sle);
		l->cache_len++;
	object_unlock(l);
} /* link_cache_put */


packet_t *link_get(link_ctx *l) {
	packet_t *p;
	
	object_lock(l);
		if ((p = SIMPLEQ_FIRST(&l->in)) != NULL) {
			SIMPLEQ_REMOVE_HEAD(&l->in,p,sqe);
			l->in_len--;
		}
	object_unlock(l);

	return p;
} /* link_get */


void link_put(link_ctx *l,packet_t *p) {
	ASSERT(p != NULL,"null packet passed to %s()",__func__);

	object_lock(l);
		FLAG_XSET(p->data[0],PACKET_LINK_NBND);
		SIMPLEQ_INSERT_TAIL(&l->out,p,sqe);
		l->out_len++;
	object_unlock(l);
} /* link_put */


int link_read(link_ctx *l,unsigned char buf[PACKET_SIZE]) {
	packet_t *p;
	
	if ((p = link_get(l)) == NULL)
		return LINKWOULDBLOCK;

	memcpy(buf,p->data,PACKET_SIZE);
	link_cache_put(l,p);

	return 0;
} /* link_read */


int link_write(link_ctx *l,unsigned char buf[PACKET_SIZE]) {
	packet_t *p;
	
	if ((p = link_cache_get(l)) == NULL)
		return LINKWOULDBLOCK;

	memcpy(p->data,buf,PACKET_SIZE);
	link_put(l,p);

	return 0;
} /* link_write */


int link_initiate(link_ctx *l) {
	unsigned char buf[(8+KE_PRIME_SIZE+KE_PUBKEY_SIZE+KE_SEED_SIZE)] =
		{'A','n','o','n','N','e','t',PROTOCOL_MAJOR,0};
	unsigned char *b; unsigned char *buf_p; unsigned char *buf_y;
	unsigned char *buf_z; unsigned char *buf_s;
	dh_ctx dh; mpz_t y;
	unsigned char kdf_s[KDF_SEED_SIZE];
	unsigned char pk[PACKET_SIZE]; int n;

	/* buf should be initialized completely (by the final char, 0) */
	buf_p	= buf_z = buf + 8;
	buf_y	= buf_p	+ KE_PRIME_SIZE;
	buf_s	= buf_y + KE_PUBKEY_SIZE;


	dh_init(&dh);
	mpz_init(y);


	/*
	 * generate ephemeral dh parameters and kdf seed
	 */
	dh_generate_parameters(&dh,KE_PRIME_BIT_SIZE,KE_GENERATOR);
	dh_generate_key(&dh);

	rand_get(kdf_s,KE_SEED_SIZE);

	bignum_format_u(dh.p,buf_p);
	bignum_format_u(dh.y,buf_y);

	memcpy(buf_s,kdf_s,KE_SEED_SIZE);


	/*
	 * send magic header, prime, pubkey and seed
	 */
	b = buf;
	n = sizeof(buf);
	if (!link_do_write(l,b,n,false))
		goto fail_early;


	/*
	 * receive magic header, pubkey and seed
	 */
	b = buf;
	n = 8 + KE_PUBKEY_SIZE + KE_SEED_SIZE;
	if (!link_do_read(l,b,n,false))
		goto fail_early;

	
	/*
	 * process response
	 */
	b = buf;
	buf_s = buf_p + KE_PRIME_SIZE;

	/* check magic header */
	if (!CHECK_MAGIC_HEADER(b))
		goto fail_early;

	bignum_parse_u(y,KE_PUBKEY_SIZE,buf_p);
	memcpy(kdf_s+KE_SEED_SIZE,buf_s,KE_SEED_SIZE);

	dh_compute_key(&dh,y);
	if (dh_check(&dh) != DH_SAFE)
		goto fail_early;
	n = bignum_format_u(dh.z,buf_z);
	
	prand_init(&l->kdf,KDF_HASH,buf_z,n,kdf_s,KDF_SEED_SIZE); 

	prand_get(&l->kdf,buf_z,256);
	prand_init(&l->arc4_out,ARCFOUR_STREAM,buf_z,256);
	PRAND_JUMP(&l->arc4_out,ARCFOUR_JUMP); /* beginning of stream is weak */

	prand_get(&l->kdf,buf_z,256);
	prand_init(&l->arc4_in,ARCFOUR_STREAM,buf_z,256);
	PRAND_JUMP(&l->arc4_in,ARCFOUR_JUMP);


	/*
	 * send hopid (we're encrypted now)
	 */
	b = buf;
	n = 8 + 1;
	*(b+8) = (l->iscaller)? l->hopid : (l->hopid + 1);
	if (!link_do_write(l,b,n,true))
		goto fail_late;


	/*
	 * receive confirmation
	 */
	 if (!link_do_read(l,b,n,true))
	 	goto fail_late;

	if (!CHECK_MAGIC_HEADER(b))
		goto fail_late;

	if (*(b+8) != (l->iscaller)? l->hopid : (l->hopid + 1))
		goto fail_late;


	mpz_clear(y);
	dh_destroy(&dh);

	object_lock(l);
	l->initiated	= true;
	l->failed	= false;
	object_unlock(l);

	bzero(buf,sizeof(buf));

	link_register(l);

	return 0;

	fail_late:

	prand_destroy(&l->kdf);
	prand_destroy(&l->arc4_in);
	prand_destroy(&l->arc4_out);

	fail_early:

	mpz_clear(y);
	dh_destroy(&dh);

	object_lock(l);
	l->initiated	= false;
	l->failed	= true;
	object_unlock(l);

	bzero(buf,sizeof(buf));

	return LINKFAILED;
} /* link_initiate */


int link_respond(link_ctx *l) {
	unsigned char buf[(8+KE_PRIME_SIZE+KE_PUBKEY_SIZE+KE_SEED_SIZE)] =
		{'A','n','o','n','N','e','t',PROTOCOL_MAJOR,0};
	unsigned char *b; unsigned char *buf_p; unsigned char *buf_y;
	unsigned char *buf_z; unsigned char *buf_s;
	dh_ctx dh; mpz_t p; mpz_t y;
	unsigned char kdf_s[KDF_SEED_SIZE];
	unsigned char pk[PACKET_SIZE]; int n;

	/* buf should be initialized completely (by the final char, 0) */
	buf_p	= buf_z = buf + 8;
	buf_y	= buf_p	+ KE_PRIME_SIZE;
	buf_s	= buf_y + KE_PUBKEY_SIZE;


	dh_init(&dh);
	mpz_init(p);
	mpz_init(y);


	/*
	 * receive magic header, prime and pubkey
	 */
	b = buf;
	n = sizeof(buf);
	if (!link_do_read(l,b,n,false))
		goto fail_early;


	/*
	 * process initiation request
	 */
	if (!CHECK_MAGIC_HEADER(buf))
		goto fail_early;

	bignum_parse_u(p,KE_PRIME_SIZE,buf_p);
	bignum_parse_u(y,KE_PUBKEY_SIZE,buf_y);
	memcpy(kdf_s,buf_s,KE_SEED_SIZE);

	dh_set_prime(&dh,p);
	dh_set_generator(&dh,KE_GENERATOR);
	dh_generate_key(&dh);
	dh_compute_key(&dh,y);

	if (dh_check(&dh) != DH_SAFE) /* check for safe parameters */
		goto fail_early;


	/*
	 * prepare to send response
	 */
	buf_y = buf_p;
	buf_s = buf_y + KE_PUBKEY_SIZE;
	
	n = bignum_format_u(dh.z,buf_z);
	rand_get(kdf_s+KE_SEED_SIZE,KE_SEED_SIZE);

	prand_init(&l->kdf,KDF_HASH,buf_z,n,kdf_s,KDF_SEED_SIZE);

	prand_get(&l->kdf,buf_z,256);
	prand_init(&l->arc4_in,ARCFOUR_STREAM,buf_z,256);
	PRAND_JUMP(&l->arc4_in,ARCFOUR_JUMP);

	prand_get(&l->kdf,buf_z,256);
	prand_init(&l->arc4_out,ARCFOUR_STREAM,buf_z,256);
	PRAND_JUMP(&l->arc4_out,ARCFOUR_JUMP);

	bzero(buf_y,KE_PUBKEY_SIZE+KE_SEED_SIZE);
	bignum_format_u(dh.y,buf_y);
	memcpy(buf_s,kdf_s+KE_SEED_SIZE,KE_SEED_SIZE);


	/*
	 * send magic header, pubkey and seed response
	 */
	b = buf;
	n = 8 + KE_PUBKEY_SIZE + KE_SEED_SIZE;
	if (!link_do_write(l,b,n,false))
		goto fail_late;


	/*
	 * receive hopid (we're encrypted now)
	 */
	b = buf;
	n = 8 + 1;
	if (!link_do_read(l,b,n,true))
		goto fail_late;

	if (!CHECK_MAGIC_HEADER(b))
		goto fail_late;

	l->hopid = *(buf+8);


	/*
	 * return to confirm
	 */
	 if (!link_do_write(l,b,n,true))
	 	goto fail_late;


	mpz_clear(p);
	mpz_clear(y);
	dh_destroy(&dh);

	object_lock(l);
	l->initiated	= true;
	l->failed	= false;
	object_unlock(l);

	bzero(buf,sizeof(buf));

	link_register(l);

	return 0;

	fail_late:

	prand_destroy(&l->kdf);
	prand_destroy(&l->arc4_in);
	prand_destroy(&l->arc4_out);

	fail_early:

	mpz_clear(p);
	mpz_clear(y);
	dh_destroy(&dh);

	object_lock(l);
	l->initiated	= false;
	l->failed	= true;
	object_unlock(l);

	bzero(buf,sizeof(buf));

	return LINKFAILED;
} /* link_respond */


channel_ctx *channel_create(link_ctx *l) {
	channel_ctx *c;
	
	if ((c = _malloc(sizeof(channel_ctx))) == NULL) {
		ERROR("could not malloc space for a new channel");
		return NULL;
	}
	bzero(c,sizeof(channel_ctx));

	object_init(c,channel_free,c);
	object_retain((c->link = l));

	c->hopid	= l->hopid;
	c->iscaller	= false;
	c->initiated	= false;
	c->secure	= false;
	c->failed	= false;
	c->next_in	= 0;
	c->next_out	= 0;


	__objects_created++;

	return c;
} /* channel_create */


void channel_free(void *c) {

	object_release(((channel_ctx *)c)->link);
	object_destroy(c);

	_free(c);

	__objects_freed++;

	return;	
} /* channel_free */


bool channel_do_read(channel_ctx *c,void *_buf,size_t len,int f) {
	unsigned char *buf = _buf;
	packet_t *pkt;

	while (len > 0) {
		get_packet: if ((pkt = channel_get(c)) == NULL) {
			object_lock(c->link);
			if (c->link->failed) {
				object_unlock(c->link);
				c->status = CHANNELFAILED;
				c->failed = true;;
				return false;
			} else if (c->link->in_len == 0) {
				object_wait(c->link);
			}
			object_unlock(c->link);
			goto get_packet;
		}

		if (f != PACKET_GET_FLAG(pkt->data,c->hopid)) {
			WARN("unexpected flag in packet");
			c->status = CHANNELFAILED;
			c->failed = true;
			return false;
		}

		if (MIN(len,PACKET_DATA_SIZE) != PACKET_GET_PLEN(pkt->data,c->hopid)) {
			WARN("unexpected payload size in packet");
			c->status = CHANNELFAILED;
			c->failed = true;
			return false;
		}

		memcpy(buf,PACKET_DATA(pkt->data,c->hopid),MIN(len,PACKET_DATA_SIZE));

		link_cache_put(c->link,pkt);

		len -= MIN(len,PACKET_DATA_SIZE);
		buf += PACKET_DATA_SIZE;
	}

	return true;
} /* channel_do_read */


bool channel_do_write(channel_ctx *c,void *_buf,size_t len,int f) {
	unsigned char *buf = _buf;
	packet_t *pkt;

	while (len > 0) {
		get_packet: if ((pkt = link_cache_get(c->link)) == NULL) {
				c->failed = true;
				c->status = CHANNELFAILED;
				return false;
		}

		PACKET_SET_FLAG(pkt->data,c->hopid,f);
		PACKET_SET_PLEN(pkt->data,c->hopid,MIN(len,PACKET_DATA_SIZE));

		memcpy(PACKET_DATA(pkt->data,c->hopid),buf,MIN(len,PACKET_DATA_SIZE));

		len -= MIN(len,PACKET_DATA_SIZE);
		buf += PACKET_DATA_SIZE;
		
		channel_put(c,pkt);
	}

	return true;
} /* channel_do_write */


packet_t *channel_get(channel_ctx *c) {
	packet_t *pkt;
	unsigned char hmac[HASH_DIGEST_SIZE_MAX];
	
	if ((pkt = link_get(c->link)) == NULL) {
		c->status = CHANNELWOULDBLOCK;
		return NULL;
	}

	/*
	 * unwrap packet
	 */
	if (c->secure) {

		/*** decrypt ***/
		cipher_decrypt(&c->cipher_in,
			PACKET_DECRYPT(pkt->data,c->hopid),
			PACKET_DECRYPT_SIZE(c->hopid));

		/*** check sequence ***/
		if (c->next_in != PACKET_GET_CSEQ(pkt->data,c->hopid)) {
			WARN("bad sequence in packet!");
			c->status = CHANNELFAILED;
			return NULL;
		}

		/*** check hmac ***/
		HMAC(TIGER_192,hmac,c->key_in,sizeof(c->key_in),
			PACKET_HASH(pkt->data,c->hopid),
			PACKET_HASH_SIZE(c->hopid));

		if (memcmp(hmac,PACKET_HMAC(pkt->data,c->hopid),PACKET_HMAC_SIZE) != 0) {
			WARN("bad MAC found in packet!");
			c->status = CHANNELFAILED;
			return NULL;
		}

		c->next_in++;
	}

	return pkt;
} /* channel_get */


void channel_put(channel_ctx *c,packet_t *pkt) {
	unsigned char hmac[HASH_DIGEST_SIZE_MAX];
	

	/*
	 * wrap packet
	 */
	if (c->secure) {

		/*** set sequence ***/
		PACKET_SET_CSEQ(pkt->data,c->hopid,c->next_out);

		/*** create hmac ***/
		HMAC(TIGER_192,hmac,c->key_out,sizeof(c->key_in),
			PACKET_HASH(pkt->data,c->hopid),
			PACKET_HASH_SIZE(c->hopid));
		memcpy(PACKET_HMAC(pkt->data,c->hopid),hmac,PACKET_HMAC_SIZE);

		/*** encrypt ***/
		cipher_encrypt(&c->cipher_out,
			PACKET_ENCRYPT(pkt->data,c->hopid),
			PACKET_ENCRYPT_SIZE(c->hopid));

		c->next_out++;
	}

	link_put(c->link,pkt);
} /* channel_put */


/*
 * initiator sends:
 * 	8 bytes magic header = 1 packet (8 bytes)
 * receiver sends:
 *	8 bytes magic header + 128 bytes prime +
 * 	128 bytes pubkye + 16 bytes seed = 6 packets (280 bytes)
 * initiator sends:
 * 	8 bytes magic header + 128 bytes pubkey +
 * 	16 bytes seed = 4 packets (152 bytes)
 */
int channel_initiate(channel_ctx *c) {
	unsigned char buf[(MAGIC_HEADER_SIZE+KE_PRIME_SIZE+
		KE_PUBKEY_SIZE+KE_SEED_SIZE)] = MAGIC_HEADER_STATIC;
	unsigned char *b; unsigned char *buf_p; unsigned char *buf_y;
	unsigned char *buf_z; unsigned char *buf_s;
	dh_ctx dh; mpz_t p; mpz_t y;
	unsigned char kdf_seed[KDF_SEED_SIZE] = {0};
	unsigned char *kdf_east, *kdf_west, *iv_in, *iv_out;
	unsigned char key[16+16+16+16] = {0};
	unsigned char *ckey_in, *ckey_out, *hkey_in, *hkey_out;
	packet_t *pkt; int n, i;


	/*
	 * initialize DH and GMP structs
	 */
	dh_init(&dh);
	mpz_init(p);
	mpz_init(y);

	/*
	 * position buffer pointers
	 */
	b		= buf;
	buf_p		= b + MAGIC_HEADER_SIZE;
	buf_y		= buf_p + KE_PRIME_SIZE;
	buf_s		= buf_y + KE_PUBKEY_SIZE;
	kdf_east	= kdf_seed;
	kdf_west	= kdf_seed + KE_SEED_SIZE;
	iv_in		= kdf_east;
	iv_out		= kdf_west;
	hkey_in		= key;
	hkey_out	= hkey_in + 16;
	ckey_in		= hkey_out + 16;
	ckey_out	= ckey_in + 16;


	/*
	 * send one packet w/ header
	 * to prove we're a proper client, then receive all the require
	 * info from the remote node
	 *
	 * NOTE: these are asynchronous, i.e. the receiver can send
	 *       its data before it receives this first packet
	 *
	 */
	if (!channel_do_write(c,b,MAGIC_HEADER_SIZE,PACKET_CHANNEL_INIT))
		goto fail_early;

	if (!channel_do_read(c,b,sizeof(buf),PACKET_CHANNEL_INIT))
		goto fail_early;


	/*
	 * now check if received header is right
	 */
	if (!CHECK_MAGIC_HEADER(b))
		goto fail_early;


	/*
	 * setup our DH parameters and kdf seed
	 */
	bignum_parse_u(p,KE_PRIME_SIZE,buf_p);
	bignum_parse_u(y,KE_PUBKEY_SIZE,buf_y);
	memcpy(kdf_west,buf_s,KE_SEED_SIZE);

	dh_set_prime(&dh,p);
	dh_set_generator(&dh,KE_GENERATOR);
	dh_generate_key(&dh); /*** do our DH operations ***/
	dh_compute_key(&dh,y); /*** generate shared secret ***/
	if (dh_check(&dh) != DH_SAFE) /*** funny business going on? ***/
		goto fail_early;


	/*
	 * key our primitives
	 */
	rand_get(kdf_east,KE_SEED_SIZE);
	n = bignum_format_u(dh.z,buf);
	prand_init(&c->kdf,KDF_HASH,buf,n,kdf_seed,KDF_SEED_SIZE);
	prand_get(&c->kdf,key,sizeof(key));
	cipher_init(&c->cipher_in,RIJNDAEL_128,ckey_in,iv_in);
	cipher_init(&c->cipher_out,RIJNDAEL_128,ckey_out,iv_out);	
	memcpy(c->key_in,hkey_in,16);
	memcpy(c->key_out,hkey_out,16);


	/*
	 * re-position buffer pointers
	 */
	b	= buf;
	buf_y	= b + MAGIC_HEADER_SIZE;
	buf_s	= buf_y + KE_PUBKEY_SIZE;


	/*
	 * prepare to send buffer
	 */
	bzero(b,sizeof(buf));
	SET_MAGIC_HEADER(b);

	bignum_format_u(dh.y,buf_y);
	memcpy(buf_s,kdf_east,KE_SEED_SIZE);

	/*** send it off ***/
	if (!channel_do_write(c,b,MAGIC_HEADER_SIZE+KE_PUBKEY_SIZE+KE_SEED_SIZE,PACKET_CHANNEL_INIT))
		goto fail_late;


	/*
	 * confirm
	 */
	c->secure = true;

	/*** send ours ***/
	SET_MAGIC_HEADER(buf);
	if (!channel_do_write(c,buf,MAGIC_HEADER_SIZE,PACKET_CHANNEL_INIT))
		goto fail_late;

	/*** recieve packet ***/
	if (!channel_do_read(c,buf,MAGIC_HEADER_SIZE,PACKET_CHANNEL_INIT))
		goto fail_late;
	if (!CHECK_MAGIC_HEADER(buf))
		goto fail_late;

	c->initiated = true;

	
	mpz_clear(y);
	mpz_clear(p);
	dh_destroy(&dh);

	return 0;

	
	fail_late:

	prand_destroy(&c->kdf);
	cipher_destroy(&c->cipher_in);
	cipher_destroy(&c->cipher_out);
	
	fail_early:

	mpz_clear(y);
	mpz_clear(p);
	dh_destroy(&dh);

	c->failed = true;
	c->status = CHANNELFAILED;

	return CHANNELFAILED;
} /* channel_initiate */


int channel_respond(channel_ctx *c) {
	unsigned char buf[(MAGIC_HEADER_SIZE+KE_PRIME_SIZE+
		KE_PUBKEY_SIZE+KE_SEED_SIZE)] = MAGIC_HEADER_STATIC;
	unsigned char *b; unsigned char *buf_p; unsigned char *buf_y;
	unsigned char *buf_z; unsigned char *buf_s;
	dh_ctx dh; mpz_t y;
	unsigned char kdf_seed[KDF_SEED_SIZE] = {0};
	unsigned char *kdf_east, *kdf_west, *iv_in, *iv_out;
	unsigned char key[16+16+16+16] = {0};
	unsigned char *ckey_in, *ckey_out, *hkey_in, *hkey_out;
	packet_t *pkt; int n, i;


	/*
	 * initialize DH and GMP structs
	 */
	dh_init(&dh);
	mpz_init(y);


	/*
	 * position buffer pointers
	 *
	 * NOTE: kdf_seed, iv, hmac keys and cipher keys
	 *       reversed from channe_initiate
	 */
	b		= buf;
	buf_p		= b + MAGIC_HEADER_SIZE;
	buf_y		= buf_p + KE_PRIME_SIZE;
	buf_s		= buf_y + KE_PUBKEY_SIZE;
	kdf_east	= kdf_seed;
	kdf_west	= kdf_seed + KE_SEED_SIZE;
	iv_in		= kdf_west;
	iv_out		= kdf_east;
	hkey_out	= key;
	hkey_in		= hkey_out + 16;
	ckey_out	= hkey_in + 16;
	ckey_in		= ckey_out + 16;


	/*
	 * setup our DH parameters and kdf seed
	 */
	dh_generate_parameters(&dh,KE_PRIME_BIT_SIZE,KE_GENERATOR);
	dh_generate_key(&dh);
	rand_get(kdf_west,KE_SEED_SIZE);

	bignum_format_u(dh.p,buf_p);
	bignum_format_u(dh.y,buf_y);
	memcpy(buf_s,kdf_west,KE_SEED_SIZE);


	/*
	 * our prime key, pubkey and seed and receive the intiators
	 * packet (which should already be queued)
	 *
	 * NOTE: these are asynchronous, i.e. the receiver can send
	 *       its data before it receives initiator's first packet
	 *
	 */
	if (!channel_do_write(c,b,sizeof(buf),PACKET_CHANNEL_INIT))
		goto fail_early;

	if (!channel_do_read(c,b,MAGIC_HEADER_SIZE,PACKET_CHANNEL_INIT))
		goto fail_early;



	/*
	 * now check if received header is right
	 */
	if (!CHECK_MAGIC_HEADER(b))
		goto fail_early;


	/*
	 * re-position buffer pointers
	 */
	b	= buf;
	buf_y	= b + MAGIC_HEADER_SIZE;
	buf_s	= buf_y + KE_PUBKEY_SIZE;


	/*
	 * receive initiators pubkey and seed
	 */
	if (!channel_do_read(c,b,MAGIC_HEADER_SIZE+KE_PUBKEY_SIZE+KE_SEED_SIZE,PACKET_CHANNEL_INIT))
		goto fail_early;


	/*
	 * now check if received header is right
	 */
	if (!CHECK_MAGIC_HEADER(b))
		goto fail_early;


	/*
	 * complete DH and seed
	 */
	memcpy(kdf_east,buf_s,KE_SEED_SIZE);
	bignum_parse_u(y,KE_PUBKEY_SIZE,buf_y);
	dh_compute_key(&dh,y);
	if (dh_check(&dh) != DH_SAFE) /*** funny business? ***/
		goto fail_early;


	/*
	 * key our primitives
	 */
	n = bignum_format_u(dh.z,buf);
	prand_init(&c->kdf,KDF_HASH,buf,n,kdf_seed,KDF_SEED_SIZE);
	prand_get(&c->kdf,key,sizeof(key));
	cipher_init(&c->cipher_in,RIJNDAEL_128,ckey_in,iv_in);
	cipher_init(&c->cipher_out,RIJNDAEL_128,ckey_out,iv_out);	
	memcpy(c->key_in,hkey_in,16);
	memcpy(c->key_out,hkey_out,16);


	/*
	 * confirm channel
	 */
	c->secure = true;
	 
	/*** recieve packet (which trailed initiators pubkey and seed ***/
	if (!channel_do_read(c,buf,MAGIC_HEADER_SIZE,PACKET_CHANNEL_INIT))
		goto fail_late;
	if (!CHECK_MAGIC_HEADER(buf))
		goto fail_late;

	/*** send ours ***/
	if (!channel_do_write(c,buf,MAGIC_HEADER_SIZE,PACKET_CHANNEL_INIT))
		goto fail_late;

	c->initiated = true;

	
	mpz_clear(y);
	dh_destroy(&dh);

	return 0;

	
	fail_late:

	prand_destroy(&c->kdf);
	cipher_destroy(&c->cipher_in);
	cipher_destroy(&c->cipher_out);
	
	fail_early:

	mpz_clear(y);
	dh_destroy(&dh);

	c->failed = true;
	c->status = CHANNELFAILED;

	return CHANNELFAILED;

} /* channel_respond */
