/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>

#include "common.h"
#include "log.h"
#include "thread.h"
#include "list.h"
#include "queue.h"


list_t *list_create(bool of_objects) {
	list_t *list;
	
	if ((list = _malloc(sizeof(list_t))) == NULL) {
		WARN("Unable to create new list.");
		return NULL;
	}
	bzero(list,sizeof(list_t));

	object_init(list,list_free,list);

	list->head = NULL;
	list->tail = NULL;
	list->__of_objects = of_objects;

	return list;
}

void list_free(void *list) {
	list_item_t *item;

	object_destroy(list);

	if ((item = ((list_t *)list)->head) != NULL) {
		for (;;) { /* broken when no item->next */
			if (((list_t *)list)->__of_objects)
				object_release(item->value);

			if (item->next != NULL) {
				item = item->next;
				_free(item->prev);
			} else {
				_free(item);
				break;
			}
		} /* end infinte loop */
	} /* end first test */

	_free(list);
}

void *list_push(list_t *list,void *value) {
	list_item_t *list_item;
	
	if ((list_item = _malloc(sizeof(list_item_t))) == NULL) {
		return NULL;
	}

	list_item->prev	= list->tail;
	list_item->next	= NULL;	

	list->tail	= list_item;
	if (list->head == NULL)
		list->head = list->tail;

	list->size++;

	list_item->value = value;

	return value;
}

void *list_pop(list_t *list) {
	void *value;

	if (list->tail == NULL)
		return NULL;

	value = list->tail->value;

	if ((list->tail = list->tail->prev) == NULL) {
		_free(list->head);
		list->head = NULL;
		list->tail = NULL;
	} else {
		_free(list->tail->next);
		list->tail->next = NULL;
	}

	list->size--;

	return value;
}

void *list_unshift(list_t *list,void *value) {
	list_item_t *list_item;
	
	if ((list_item = _malloc(sizeof(list_item_t))) == NULL) {
		return NULL;
	}

	list_item->prev = NULL;
	list_item->next = list->head;
	list->head = list_item;
	list->size++;

	list_item->value = value;

	return value;
}

void *list_shift(list_t *list) {
	void *value;
	
	if (list->head == NULL)
		return NULL;

	value = list->head->value;

	if ((list->head = list->head->next) == NULL) {
		_free(list->tail);
		list->head = NULL;
		list->tail = NULL;
	} else {
		_free(list->head->prev);
		list->head->prev = NULL;
	}

	list->size--;

	return value;
}
