/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <netinet/in.h>

#ifdef __USE_MCRYPT
#ifndef _MCRYPT_H
#define _MCRYPT_H
#include <mcrypt.h>
#endif
#endif

#include "common.h"
#include "thread.h"
#include "log.h"
#include "crypt.h"
#include "bignum.h"
#include "packet.h"
#include "node.h"


node_t *node_create(int sd) {
	node_t *node;

	if ((node = _malloc(sizeof(node_t))) == NULL) {
		ERROR("could not malloc space for a new node");
		return NULL;
	}
	bzero(node,sizeof(node_t));
	
	object_init(node,node_free,node);
	thread_init(node,node_run,node);

	node->east	= NULL;
	node->west	= NULL;
	node->channel	= NULL;

	if ((node->east = link_create(sd)) == NULL) {
		node_free(node);
		return NULL;
	}

	__objects_created++;

	return node;
} /* end node_create */


void node_free(void *o) {
	node_t *n = (node_t *)o;

	thread_destroy(n);
	object_destroy(n);

	if (n->east != NULL)
		link_free(n->east);
	if (n->west != NULL)
		link_free(n->west);
	if (n->channel != NULL)
		channel_free(n->channel);

	_free(n);

	__objects_freed++;
	
	return; /* void */
} /* end node_free */


void node_call_back(int sd) {
	node_t *node;
	
	if ((node = node_create(sd)) == NULL) {
		ERROR("unable to create node");
		return; /* void */
	}

	thread_run(node);
	
	return; /* void */

} /* node_call_back */


void node_run(void *o) {
	node_t *n	= (node_t *)o;	
	link_ctx *l	= NULL;
	channel_ctx *c	= NULL;

	DEBUG("node started");


	/*
	 * establish link
	 */
	link_respond(n->east);
	
	if (n->east->failed)
		goto fail;

	DEBUG("link established");


	/*
	 * establish channel
	 */
	if ((n->channel = channel_create(n->east)) == NULL)
		goto fail;

	object_retain(n->channel);
	channel_respond(n->channel);

	if (n->channel->failed)
		goto fail;
	
	DEBUG("channel established");

	object_release(n->channel);

	return;

	fail:

	DEBUG("node failed");
	object_release(n->channel);

	return; /* void */
	
} /* node_run */


