/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */
#include <signal.h>
#include <pthread.h>

#include "common.h"

#ifndef __OBJECT_H
#define __OBJECT_H


typedef void (*destructor_t)(void *);
typedef void *destructor_arg_t;
typedef struct object object_t;


struct object {
	long int	object_id;

	int		__gc_ref_count;

	destructor_t		__free;
	destructor_arg_t	__free_arg;

	pthread_mutex_t		__sync_mutex;
	pthread_mutexattr_t	__sync_mutexattr;

	pthread_mutex_t		__gc_mutex;
	pthread_mutexattr_t	__gc_mutexattr;

	pthread_mutex_t		__cond_mutex;
	pthread_mutexattr_t	__cond_mutexattr;
	pthread_cond_t		__cond_var;
	pthread_condattr_t	__cond_attr;

};

#ifdef __ENABLE_DEBUG
void object_init(void *,void (*)(void *),void *,char *);
#define object_init(o,f,a) object_init((o),(f),(a),__func__)
#else
void object_init(void *,void (*)(void *),void *);
#endif /* object_init */

#ifdef __ENABLE_DEBUG
void object_destroy(void *,char *);
#define object_destroy(o) object_destroy((o),__func__)
#else
void object_destroy(void *);
#endif /* object_destroy */

#ifdef __ENABLE_DEBUG
object_t *object_create(char *);
#define object_create() object_create(__func__)
#else
object_t *object_create(void);
#endif /* object_create */

void object_free(void *);

#ifdef __ENABLE_DEBUG
void object_lock(void *,char *);
#define object_lock(o) object_lock((o),__func__)
#else
void object_lock(void *);
#endif /* object_lock */

#ifdef __ENABLE_DEBUG
void object_unlock(void *,char *);
#define object_unlock(o) object_unlock((o),__func__)
#else
void object_unlock(void *);
#endif /* object_unlock */

#ifdef __ENABLE_DEBUG
void object_retain(void *,char *);
#define object_retain(o) object_retain((o),__func__)
#else
void object_retain(void *);
#endif /* object_retain */

#ifdef __ENABLE_DEBUG
void object_release(void *,char *);
#define object_release(o) object_release((o),__func__)
#else
void object_release(void *);
#endif /* object_release */

void object_wait(void *);
void object_timedwait(void *,const struct timespec *);
void object_signal(void *);
void object_broadcast(void *);

sig_atomic_t __objects_initialized;
sig_atomic_t __objects_created;
sig_atomic_t __objects_destroyed;
sig_atomic_t __objects_freed;
void object_tracker(void);

#endif
